# Import modules
import datetime as dt
import numpy as np
import pandas as pd
import sqlalchemy

# Import custom modules
from etl_tools.sql import sql_exec_stmt, sql_read_data, sql_upload_data, sql_copy_data
from etl_tools.aws import dynamodb_read_data, s3_upload_csv


class ExtractDeleteAndLoad(object):
    """
    Class with routines to read, delete and/or upload data from/to a database or data storage.
    """

    def __init__(
        self,
        config_dict={},
        conn_dict={},
        sqlalchemy_dict={},
    ):
        """
        Class constructor.

        Parameters:

        config_dict : dict. Configuration dictionary with connection and data parameters.
        conn_dict : dict. Connection dictionary with connection parameters.
        sqlalchemy_dict : dict. Dictionary with sqlalchemy data types.
        """

        ## Set class parameters

        # Set configuration parameters
        self.connections_dict = conn_dict
        self.configs_dict = config_dict
        self.sqlalchemy_dtypes = sqlalchemy_dict
        # Set processes names
        processes_list = ["download", "delete", "truncate", "upload"]
        # Set connection parameters
        self.conn_info_dict = {key: {} for key in processes_list}
        self.conn_suff_dict = {key: {} for key in processes_list}
        self.conn_type_dict = {key: {} for key in processes_list}
        # Iterate over download connections
        for p_name in processes_list:
            for key in self.configs_dict[f"{p_name}_connections_dict"].keys():
                # Get connection suffix
                self.conn_suff_dict[p_name][key] = self.configs_dict[
                    f"{p_name}_connections_dict"
                ][key].split("_")[-1]
                # Get connection type
                self.conn_type_dict[p_name][key] = self.configs_dict[
                    f"{p_name}_connections_dict"
                ][key].split("_")[0]
                # Get connection dictionary
                self.conn_info_dict[p_name][key] = {
                    "myoracle_client_dir": (
                        self.connections_dict["myoracle_client_dir"]
                        if "myoracle_client_dir" in self.connections_dict.keys()
                        else ""
                    ),
                    "myserver": (
                        self.connections_dict[
                            f"myserver_{self.conn_type_dict[p_name][key]}_{self.conn_suff_dict[p_name][key]}"
                        ]
                        if f"myserver_{self.conn_type_dict[p_name][key]}_{self.conn_suff_dict[p_name][key]}"
                        in self.connections_dict.keys()
                        else ""
                    ),
                    "mydatabase": (
                        self.connections_dict[
                            f"mydatabase_{self.conn_type_dict[p_name][key]}_{self.conn_suff_dict[p_name][key]}"
                        ]
                        if f"mydatabase_{self.conn_type_dict[p_name][key]}_{self.conn_suff_dict[p_name][key]}"
                        in self.connections_dict.keys()
                        else ""
                    ),
                    "myusername": (
                        self.connections_dict[
                            f"myusername_{self.conn_type_dict[p_name][key]}_{self.conn_suff_dict[p_name][key]}"
                        ]
                        if f"myusername_{self.conn_type_dict[p_name][key]}_{self.conn_suff_dict[p_name][key]}"
                        in self.connections_dict.keys()
                        else ""
                    ),
                    "mypassword": (
                        self.connections_dict[
                            f"mypassword_{self.conn_type_dict[p_name][key]}_{self.conn_suff_dict[p_name][key]}"
                        ]
                        if f"mypassword_{self.conn_type_dict[p_name][key]}_{self.conn_suff_dict[p_name][key]}"
                        in self.connections_dict.keys()
                        else ""
                    ),
                    "mycharset": (
                        self.connections_dict[
                            f"mycharset_{self.conn_type_dict[p_name][key]}_{self.conn_suff_dict[p_name][key]}"
                        ]
                        if f"mycharset_{self.conn_type_dict[p_name][key]}_{self.conn_suff_dict[p_name][key]}"
                        in self.connections_dict.keys()
                        else ""
                    ),
                    "myencoding": (
                        self.connections_dict[
                            f"awmyencoding_{self.conn_type_dict[p_name][key]}_{self.conn_suff_dict[p_name][key]}"
                        ]
                        if f"awmyencoding_{self.conn_type_dict[p_name][key]}_{self.conn_suff_dict[p_name][key]}"
                        in self.connections_dict.keys()
                        else ""
                    ),
                }

    def delete_data(self, **kwargs):
        """
        Function to delete data from the database.

        Parameters:

        kwargs : dict. Keyword arguments to pass to the delete statement.
        """

        # Set keyword arguments as variables
        for key, val in kwargs.items():
            locals()[key] = val
        # Iterate over delete statements
        for key, stmt in self.configs_dict["delete_sql_stmts_dict"].items():
            # Start date to ingest data from
            global start_date
            start_date = (
                eval(self.configs_dict["delete_dates_dict"][key]["start_date"])
                if "eval(" in self.configs_dict["delete_dates_dict"][key]["start_date"]
                else self.configs_dict["delete_dates_dict"][key]["start_date"]
            )
            # End date to ingest data from
            global end_date
            end_date = (
                eval(self.configs_dict["delete_dates_dict"][key]["end_date"])
                if "eval(" in self.configs_dict["delete_dates_dict"][key]["end_date"]
                else self.configs_dict["delete_dates_dict"][key]["end_date"]
            )
            print(f"Deleting data from {key}, since {start_date} to {end_date}...")
            # Get connection suffix
            conn_suff = self.conn_suff_dict["delete"][key]
            # Get connection type
            conn_type = self.conn_type_dict["delete"][key]
            # Get connection dictionary
            conn_dict = self.conn_info_dict["delete"][key]
            # Execute delete statement
            print(f"Deleting data for {key}...")
            try:
                sql_exec_stmt(
                    eval(stmt) if "{" in stmt else stmt, conn_dict, mode=conn_type
                )
            except Exception as e:
                print(f"Error deleting data: {type(e)} - {e}")

        pass

    def truncate_data(self, **kwargs):
        """
        Function to truncate data from the database.
        """

        # Iterate over truncate statements
        for key, stmt in self.configs_dict["truncate_sql_stmts_dict"].items():
            # Get connection suffix
            conn_suff = self.conn_suff_dict["truncate"][key]
            # Get connection type
            conn_type = self.conn_type_dict["truncate"][key]
            # Get connection dictionary
            conn_dict = self.conn_info_dict["truncate"][key]
            # Execute truncate statement
            print(f"Truncating data for {key}...")
            try:
                sql_exec_stmt(
                    eval(stmt) if "{" in stmt else stmt, conn_dict, mode=conn_type
                )
            except Exception as e:
                print(f"Error truncating data: {type(e)} - {e}")

        pass

    def read_data(self):
        """
        Function to read CMV data from DynamoDB.
        """

        # Initialize raw data
        self.raw_data = {}
        # Iterate over tables/statements
        for key, tb_name in self.configs_dict["download_table_names_dict"].items():
            # Start date to ingest data from
            global start_date
            start_date = (
                eval(self.configs_dict["download_dates_dict"][key]["start_date"])
                if "eval("
                in self.configs_dict["download_dates_dict"][key]["start_date"]
                else self.configs_dict["download_dates_dict"][key]["start_date"]
            )
            # End date to ingest data from
            global end_date
            end_date = (
                eval(self.configs_dict["download_dates_dict"][key]["end_date"])
                if "eval(" in self.configs_dict["download_dates_dict"][key]["end_date"]
                else self.configs_dict["download_dates_dict"][key]["end_date"]
            )
            print(f"Reading data from {key}, since {start_date} to {end_date}...")
            # Get connection suffix
            conn_suff = self.conn_suff_dict["download"][key]
            # Get connection type
            conn_type = self.conn_type_dict["download"][key]
            # Get connection dictionary
            conn_dict = self.conn_info_dict["download"][key]
            # Evaluate keyword arguments
            kwargs_eval = {
                kw_key: eval(kw_val) if "{" in kw_val else kw_val
                for kw_key, kw_val in self.configs_dict["download_kwargs_dict"][
                    key
                ].items()
            }
            # Read data
            if conn_type == "dynamodb":
                data = dynamodb_read_data(
                    tb_name,
                    self.connections_dict[f"aws_access_key_id_{conn_suff}"],
                    self.connections_dict[f"aws_secret_access_key_{conn_suff}"],
                    self.connections_dict[f"region_name_{conn_suff}"],
                    **kwargs_eval,
                )
            else:
                data = sql_read_data(
                    (
                        eval(self.configs_dict["download_sql_stmts_dict"][key])
                        if "{" in self.configs_dict["download_sql_stmts_dict"][key]
                        else self.configs_dict["download_sql_stmts_dict"][key]
                    ),
                    conn_dict,
                    mode=conn_type,
                    custom_conn_str=self.configs_dict["download_custom_conn_strs_dict"][
                        key
                    ],
                    connect_args=self.configs_dict["download_connect_args_dict"][key],
                    name=tb_name,
                    max_n_try=3,
                )
            # Add data to raw data dictionary
            self.raw_data[key] = data.copy()
        # Free memory
        del data

        pass

    def upload_data(self, data_to_upload: dict):
        """
        Function to upload CMV data into the database.

        Parameters:

        data_to_upload : list. List with data to upload.
        """

        # Iterate over tables/statements
        for key, tb_name in self.configs_dict["upload_tables_dict"].items():
            # Set data to upload
            upload_data = data_to_upload[key]
            # Get connection suffix
            conn_suff = self.conn_suff_dict["upload"][key]
            # Get connection type
            conn_type = self.conn_type_dict["upload"][key]
            # Get connection dictionary
            conn_dict = self.conn_info_dict["upload"][key]
            # Upload data
            if conn_type == "dynamodb":
                pass
            else:
                # Define column names and data types
                col_dict = self.configs_dict["upload_python_to_sql_dtypes_dict"][key]
                # Define dictionary with data types
                dtypes_dict = {
                    col: (
                        eval(
                            f'{self.sqlalchemy_dtypes[col_dtype.split("(")[0]]}({col_dtype.split("(")[1][:-1]})'
                        )
                        if "(" in col_dtype
                        else eval(f"{self.sqlalchemy_dtypes[col_dtype]}()")
                    )
                    for col, col_dtype in col_dict.items()
                }
                # Use order defined in data types dictionary
                upload_data = upload_data[list(col_dict.keys())]
                # Upload data to created tables
                if "redshift" in conn_type:
                    # Upload data to S3
                    s3_upload_csv(
                        upload_data,
                        self.configs_dict["s3_file_paths_dict"][key],
                        self.connections_dict[f"aws_access_key_id_{conn_suff}"],
                        self.connections_dict[f"aws_secret_access_key_{conn_suff}"],
                        region_name=self.connections_dict[f"region_name_{conn_suff}"],
                        sep=",",
                        index=False,
                        encoding="utf-8",
                    )
                    # Copy data from S3 to database
                    sql_copy_data(
                        (
                            eval(self.configs_dict["s3_file_paths_dict"][key])
                            if "{" in self.configs_dict["s3_file_paths_dict"][key]
                            else self.configs_dict["s3_file_paths_dict"][key]
                        ),
                        self.configs_dict["upload_schemas_dict"][key],
                        self.configs_dict["upload_tables_dict"][key],
                        conn_dict,
                        self.connections_dict[f"aws_access_key_id_aws_{conn_suff}"],
                        self.connections_dict[f"aws_secret_access_key_aws_{conn_suff}"],
                        self.connections_dict[f"region_name_aws_{conn_suff}"],
                        name=self.configs_dict["upload_tables_dict"][key],
                        max_n_try=self.configs_dict["max_n_try"],
                    )
                else:
                    # Upload data to database
                    sql_upload_data(
                        upload_data,
                        self.configs_dict["upload_schemas_dict"][key],
                        self.configs_dict["upload_tables_dict"][key],
                        conn_dict,
                        mode=conn_type,
                        custom_conn_str=self.configs_dict[
                            "upload_custom_conn_strs_dict"
                        ][key],
                        connect_args=self.configs_dict["upload_connect_args_dict"][key],
                        name=self.configs_dict["upload_tables_dict"][key],
                        chunksize=self.configs_dict["upload_chunksizes_dict"][key],
                        method=self.configs_dict["upload_methods_dict"][key],
                        dtypes_dict=dtypes_dict,
                        max_n_try=self.configs_dict["max_n_try"],
                        n_jobs=self.configs_dict["n_parallel"],
                    )

        pass



# Import modules
import numpy as np
import pandas as pd
import json
import pickle
import awswrangler as aws
import boto3
import datetime as dt


def s3_list_objects(
    s3_bucket_name,
    s3_path,
    aws_access_key,
    aws_secret_access_key,
    region_name="us-east-1",
):
    """
    Function to list objects from S3 bucket.

    s3_bucket_name:                 str. Name of the S3 bucket without "s3://" prefix.
    s3_path:                        str. Path to the file in the S3 bucket (relative to root).
    aws_access_key:                 str. Name of the environment variable with the AWS access key.
    aws_secret_access_key:          str. Name of the environment variable with the AWS secret access key.
    region_name:                    str. Name of the AWS region to use.
    """

    # Create S3 session
    my_session = boto3.Session(
        region_name=region_name,
        aws_access_key_id=aws_access_key,
        aws_secret_access_key=aws_secret_access_key,
    )
    # Set up S3 resource
    s3 = my_session.client("s3")
    # Get objects list
    objects_list = [
        obj_dict["Key"]
        for obj_dict in s3.list_objects_v2(Bucket=s3_bucket_name, Prefix=s3_path)[
            "Contents"
        ]
    ]
    # Check if limit of 1000 objects was reached
    if len(objects_list) == 1000:
        # Get paginator
        paginator = s3.get_paginator("list_objects_v2")
        # Get pages
        pages = paginator.paginate(Bucket=s3_bucket_name, Prefix=s3_path)
        # Iterate over pages
        objects_list = np.hstack(
            [[obj_dict["Key"] for obj_dict in page["Contents"]] for page in pages]
        )

    return objects_list


def s3_get_object(
    s3_bucket_name,
    s3_path,
    aws_access_key,
    aws_secret_access_key,
    region_name="us-east-1",
):
    """
    Function to get object from S3 bucket.

    s3_bucket_name:                 str. Name of the S3 bucket without "s3://" prefix.
    s3_path:                        str. Path to the file in the S3 bucket (relative to root).
    aws_access_key:                 str. Name of the environment variable with the AWS access key.
    aws_secret_access_key:          str. Name of the environment variable with the AWS secret access key.
    region_name:                    str. Name of the AWS region to use.
    """

    # Create S3 session
    my_session = boto3.Session(
        region_name=region_name,
        aws_access_key_id=aws_access_key,
        aws_secret_access_key=aws_secret_access_key,
    )
    # Set up S3 resource
    s3 = my_session.client("s3")
    # Set up S3 object
    content_object = s3.get_object(Bucket=s3_bucket_name, Key=s3_path)

    return content_object


def s3_put_object(
    s3_body_content,
    s3_bucket_name,
    s3_path,
    aws_access_key,
    aws_secret_access_key,
    region_name="us-east-1",
):
    """
    Function to put object on S3 bucket.

    s3_body_content:                bytes. Content to be uploaded to S3.
    s3_bucket_name:                 str. Name of the S3 bucket without "s3://" prefix.
    s3_path:                        str. Path to the file in the S3 bucket (relative to root).
    aws_access_key:                 str. Name of the environment variable with the AWS access key.
    aws_secret_access_key:          str. Name of the environment variable with the AWS secret access key.
    region_name:                    str. Name of the AWS region to use.
    """

    # Create S3 session
    my_session = boto3.Session(
        region_name=region_name,
        aws_access_key_id=aws_access_key,
        aws_secret_access_key=aws_secret_access_key,
    )
    # Set up S3 resource
    s3 = my_session.client("s3")
    # Set up S3 object
    s3.put_object(Body=s3_body_content, Bucket=s3_bucket_name, Key=s3_path)

    pass


def s3_read_file(
    s3_bucket_name,
    s3_path,
    aws_access_key,
    aws_secret_access_key,
    region_name="us-east-1",
    encoding="utf-8",
    file_type="plain",
):
    """
    Function to read .csv or .json file from S3 bucket.

    s3_bucket_name:                 str. Name of the S3 bucket without "s3://" prefix.
    s3_path:                        str. Path to the file in the S3 bucket (relative to root).
    aws_access_key:                 str. Name of the environment variable with the AWS access key.
    aws_secret_access_key:          str. Name of the environment variable with the AWS secret access key.
    region_name:                    str. Name of the AWS region to use.
    encoding:                       str. Encoding to use for reading the file.
    file_type:                      str. Type of file to read ("csv" or "plain"). Default "plain".
    """

    # Get content object
    content_object = s3_get_object(
        s3_bucket_name,
        s3_path,
        aws_access_key,
        aws_secret_access_key,
        region_name=region_name,
    )
    if file_type == "csv":
        # Get and decode content
        file_content = content_object.get("Body")
    elif file_type == "plain":
        # Get and decode content
        file_content = content_object.get("Body").read().decode(encoding)

    return file_content


def s3_read_json(
    s3_bucket_name,
    s3_path,
    aws_access_key,
    aws_secret_access_key,
    region_name="us-east-1",
    encoding="utf-8",
):
    """
    Function to read json from S3 bucket.

    s3_bucket_name:                 str. Name of the S3 bucket without "s3://" prefix.
    s3_path:                        str. Path to the json file in the S3 bucket (relative to root).
    aws_access_key:                 str. Name of the environment variable with the AWS access key.
    aws_secret_access_key:          str. Name of the environment variable with the AWS secret access key.
    region_name:                    str. Name of the AWS region to use.
    """

    # Read S3 file
    file_content = s3_read_file(
        s3_bucket_name,
        s3_path,
        aws_access_key,
        aws_secret_access_key,
        region_name=region_name,
        encoding=encoding,
        file_type="plain",
    )
    # Read json
    json_content = json.loads(file_content)

    return json_content


def s3_write_json(
    json_data,
    s3_bucket_name,
    s3_path,
    aws_access_key,
    aws_secret_access_key,
    region_name="us-east-1",
    encoding="utf-8",
):
    """
    Function to write json to S3 bucket.

    json_data:                      dict. Data to be written to json file.
    s3_bucket_name:                 str. Name of the S3 bucket without "s3://" prefix.
    s3_path:                        str. Path to the json file in the S3 bucket (relative to root).
    aws_access_key:                 str. Name of the environment variable with the AWS access key.
    aws_secret_access_key:          str. Name of the environment variable with the AWS secret access key.
    region_name:                    str. Name of the AWS region to use.
    """

    # Set and encode data
    s3_body_content = bytes(json.dumps(json_data).encode(encoding))
    # Upload content to S3
    s3_put_object(
        s3_body_content,
        s3_bucket_name,
        s3_path,
        aws_access_key,
        aws_secret_access_key,
        region_name=region_name,
    )

    pass


def s3_read_csv(
    s3_bucket_name,
    s3_path,
    aws_access_key,
    aws_secret_access_key,
    region_name="us-east-1",
    **kwargs,
):
    """
    Function to read csv from S3 bucket.

    s3_bucket_name:                 str. Name of the S3 bucket without "s3://" prefix.
    s3_path:                        str. Path to the csv file in the S3 bucket (relative to root).
    aws_access_key:                 str. Name of the environment variable with the AWS access key.
    aws_secret_access_key:          str. Name of the environment variable with the AWS secret access key.
    region_name:                    str. Name of the AWS region to use.
    kwargs:                         dict. Keyword arguments to pass to pd.read_csv.
    """

    # Read S3 file
    file_content = s3_read_file(
        s3_bucket_name,
        s3_path,
        aws_access_key,
        aws_secret_access_key,
        region_name=region_name,
        encoding="utf-8",
        file_type="csv",
    )
    # Read csv
    csv_content = pd.read_csv(file_content, **kwargs)

    return csv_content


def s3_write_parquet(
    data,
    s3_bucket_name,
    s3_path,
    aws_access_key,
    aws_secret_access_key,
    region_name="us-east-1",
):
    """
    Function to write DataFrame to .parquet in S3 bucket.

    data:                           pd.DataFrame. Data to be written to .parquet file.
    s3_bucket_name:                 str. Name of the S3 bucket without "s3://" prefix.
    s3_path:                        str. Path to the .parquet file in the S3 bucket (relative to root).
    aws_access_key:                 str. Name of the environment variable with the AWS access key.
    aws_secret_access_key:          str. Name of the environment variable with the AWS secret access key.
    region_name:                    str. Name of the AWS region to use.
    """

    # Set and encode data
    s3_body_content = data.to_parquet()
    # Upload content to S3
    s3_put_object(
        s3_body_content,
        s3_bucket_name,
        s3_path,
        aws_access_key,
        aws_secret_access_key,
        region_name=region_name,
    )
    pass


def s3_read_pkl(
    s3_bucket_name,
    s3_pickle_path,
    aws_access_key,
    aws_secret_access_key,
    region_name="us-east-1",
):
    """
    Function to read pickle file from S3.

    s3_bucket_name:                 str. Name of the S3 bucket without "s3://" prefix.
    s3_pickle_path:                 str. Path to the pickle file in the S3 bucket (relative to bucket).
    aws_access_key:                 str. Name of the environment variable with the AWS access key.
    aws_secret_access_key:          str. Name of the environment variable with the AWS secret access key.
    region_name:                    str. Name of the AWS region to use.
    """
    # Download from S3
    pkl_file = (
        s3_get_object(
            s3_bucket_name,
            s3_pickle_path,
            aws_access_key,
            aws_secret_access_key,
            region_name=region_name,
        )
        .get("Body")
        .read()
    )
    # Load threshold for center
    pickle_object = pickle.loads(pkl_file)

    return pickle_object


def s3_upload_csv(
    data,
    s3_file_path,
    aws_access_key,
    aws_secret_access_key,
    region_name="us-east-1",
    sep=",",
    index=False,
    encoding="utf-8",
):
    """
    Function to upload data as CSV to S3 bucket.

    data:                   pd.DataFrame. Data to upload.
    s3_file_path:           str. S3 file path.
    aws_access_key:         str. Name of the environment variable with the AWS access key.
    aws_secret_access_key:  str. Name of the environment variable with the AWS secret access key.
    region_name:            str. Name of the AWS region to use.
    sep:                    str. Separator to use for CSV data.
    index:                  bool. Whether to include the index in the file.
    encoding:               str. Encoding to use.
    """

    # Create S3 session
    my_session = boto3.Session(
        region_name=region_name,
        aws_access_key_id=aws_access_key,
        aws_secret_access_key=aws_secret_access_key,
    )
    # Upload data to S3 bucket
    aws.s3.to_csv(
        data,
        path=s3_file_path,
        sep=sep,
        my_session=my_session,
        index=index,
        encoding=encoding,
    )

    pass


def dynamodb_read_data(
    table_name, aws_access_key_id, aws_secret_access_key, region_name, **kwargs
):
    """
    Function to read data from DynamoDB.
    """

    # Set resource
    dynamodb = boto3.resource(
        "dynamodb",
        aws_access_key_id=aws_access_key_id,
        aws_secret_access_key=aws_secret_access_key,
        region_name=region_name,
    )
    # Evaluate keyword arguments
    kwargs_eval = {key: eval(val) for key, val in kwargs.items()}
    # Get table
    table = dynamodb.Table(table_name)
    # Scan table
    scan_response = table.scan(**kwargs_eval)
    # Get data from table
    data = scan_response["Items"]
    while "LastEvaluatedKey" in scan_response:  # Iterate over each scan response
        # Scan table
        scan_response = table.scan(
            ExclusiveStartKey=scan_response["LastEvaluatedKey"], **kwargs_eval
        )
        # Add data
        data.extend(scan_response["Items"])

    return data



# Import modules
import datetime as dt
import os
import subprocess

# Import submodules
from concurrent.futures import ProcessPoolExecutor
from colorama import Fore


def mk_exec_logs(file_path, file_name, process_name, output_content):
    """
    Function to create/save execution log files.

    file_path :     String. File path to use for saving logs.
    file_name :     String. File name to use for log file.
    process_name :  String. Process name.
    output_content: String. Output content.
    """

    ## Set file name

    file_name_wext = f"{file_name}.log"

    ## Generate log file

    # Create log file if not in files
    if file_name_wext not in os.listdir(file_path):
        with open(f"{file_path}/{file_name_wext}", "w") as l_f:
            title_str = "#                    Process output                    #"
            l_f.write("#" * len(title_str) + "\n")
            l_f.write(title_str + "\n")
            l_f.write("#" * len(title_str) + "\n\n")
    # Append error info
    with open(f"{file_path}/{file_name_wext}", "a") as l_f:
        l_f.write(
            f"Date:\n\n {dt.datetime.strftime(dt.datetime.now(),'%Y/%m/%d %H:%M:%S')}\n\n"
        )
        l_f.write(f"Process name:\n\n {process_name}\n\n")
        l_f.write(f"Output:\n\n {output_content}\n\n")
        l_f.write("------------------------------------------------------\n\n")

    pass


def mk_texec_logs(file_path, file_name, time_var, time_val, obs=None):
    """
    Function to create/save log time execution files.

    file_path :     String. File path to use for saving logs.
    file_name :     String. File name to use for log file.
    time_val :     String. Time variable's value.
    time_var :      String. Time variable's name.
    """

    ## Set file name

    file_name_wmode = f"{file_name}.log"

    ## General log file

    # Create log file if not in files
    if file_name_wmode not in os.listdir(file_path):
        with open(f"{file_path}/{file_name_wmode}", "w") as l_f:
            title_str = "# Time variable          Time value          Date          Observations          #"
            l_f.write("#" * len(title_str) + "\n")
            l_f.write(title_str + "\n")
            l_f.write("#" * len(title_str) + "\n\n")
    # Append error info
    with open(f"{file_path}/{file_name_wmode}", "a") as l_f:
        l_f.write(
            f"{time_var}          {time_val}          {dt.datetime.strftime(dt.datetime.now(),'%Y/%m/%d %H:%M:%S')}          {obs}"
            + "\n"
        )

    pass


def mk_err_logs(file_path, file_name, err_var, err_desc, mode="summary"):
    """
    Function to create/save log error files.

    file_path :     String. File path to use for saving logs.
    file_name :     String. File name to use for log file.
    err_desc :      String. Error description.
    err_var :       String. Error variable name.
    """

    ## Set file name

    file_name_wmode = f"{file_name}_{mode.lower()}.log"

    ## Generate log file

    if mode.lower() == "summary":
        ## General log file

        # Create log file if not in files
        if file_name_wmode not in os.listdir(file_path):
            with open(f"{file_path}/{file_name_wmode}", "w") as l_f:
                title_str = "# Error variable          Error description          Date          #"
                l_f.write("#" * len(title_str) + "\n")
                l_f.write(title_str + "\n")
                l_f.write("#" * len(title_str) + "\n\n")
        # Append error info
        with open(f"{file_path}/{file_name_wmode}", "a") as l_f:
            l_f.write(
                f"{err_var}          {err_desc}          {dt.datetime.strftime(dt.datetime.now(),'%Y/%m/%d %H:%M:%S')}"
                + "\n"
            )

    elif mode.lower() == "detailed":
        ## Detailed log file

        # Create log file if not in files
        if file_name_wmode not in os.listdir(file_path):
            with open(f"{file_path}/{file_name_wmode}", "w") as l_f:
                title_str = "#                    Detailed error description                    #"
                l_f.write("#" * len(title_str) + "\n")
                l_f.write(title_str + "\n")
                l_f.write("#" * len(title_str) + "\n\n")
        # Append error info
        with open(f"{file_path}/{file_name_wmode}", "a") as l_f:
            l_f.write(
                f"Date:\n\n {dt.datetime.strftime(dt.datetime.now(),'%Y/%m/%d %H:%M:%S')}\n\n"
            )
            l_f.write(f"Error variable:\n\n {err_var}\n\n")
            l_f.write(f"Error description:\n\n {err_desc}\n\n")
            l_f.write("------------------------------------------------------\n\n")

    pass


def parallel_execute(applyFunc, *args, **kwargs):
    """Function to execute function parallely.

    applyFunc :        Function. Function to apply parallely.
    args      :        Iterable. Arguments to pass to function on each parallel execution.
    """

    with ProcessPoolExecutor() as executor:
        results = executor.map(applyFunc, *args, **kwargs)

    return results


def execute_script(
    process_str,
    log_file_path="logs",
    exec_log_file_name="exec.log",
    texec_log_file_name="txec.log",
):
    """
    Function to execute an script, saving execution logs.

    process_str :           String. Process to execute.
    log_file_path :         String. File path to use for saving logs.
    exec_log_file_name :    String. Execution log file name.
    texec_log_file_name :   String. Time execution log file name.
    """

    # Execute process
    s = dt.datetime.now()
    r = subprocess.check_output(
        process_str,
        shell=True,
        stderr=subprocess.STDOUT,
        universal_newlines=True,
    )
    e = dt.datetime.now()
    print(f"----- Process execution duration = {Fore.GREEN}{e-s}{Fore.RESET} -----")
    # Create execution logs
    os.makedirs(log_file_path, exist_ok=True)
    mk_exec_logs(
        log_file_path,
        exec_log_file_name,
        f"'{process_str}'",
        r,
    )
    mk_texec_logs(log_file_path, texec_log_file_name, f"'{process_str}'", e - s)

    pass
